"""Authentication backend for Ashley"""

import logging

from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied

from lti_provider.default.backend import LTIBackend as BaseLTIBackend
from lti_provider.lti import LTI

logger = logging.getLogger(__name__)


class LTIBackend(BaseLTIBackend):
    """
    Authentication backend using LTI launch request to authenticate a user.
    If a user does not exist in database, it creates it.
    The username is generated with the remote (consumer) user id and with
    the consumer identifier, to be able to authenticate users from multiple
    consumer sites without any collision.
    """

    def authenticate(self, request, username=None, password=None, **kwargs):
        lti_request: LTI = kwargs.get("lti_request")
        if not lti_request:
            return None

        if not lti_request.is_valid:
            raise PermissionDenied()

        lti_consumer = lti_request.get_consumer()
        if not lti_consumer:
            raise PermissionDenied

        email = self._get_mandatory_param(
            lti_request, "lis_person_contact_email_primary"
        )
        remote_user_id = self._get_remote_user_id(lti_request)

        logger.debug(
            "User %s (consumer = %s) authenticated from LTI request",
            username,
            lti_consumer.slug,
        )
        user_model = get_user_model()
        try:
            return user_model.objects.get(
                lti_consumer=lti_consumer, lti_remote_user_id=remote_user_id
            )
        except user_model.DoesNotExist:
            username = f"{remote_user_id}@{lti_consumer.slug:s}"

            user = user_model.objects.create_user(
                username,
                email=email,
                lti_consumer=lti_consumer,
                lti_remote_user_id=remote_user_id,
                public_username=remote_user_id,
            )
            logger.debug("User %s created in database", username)

        if not user.is_active:
            logger.debug("User %s is not active", user.username)
            raise PermissionDenied()
        return user

    @staticmethod
    def _get_remote_user_id(lti_request: LTI):
        """
        Get the remote user id.
        It can be in different LTI parameters, depends on the LTI consumer.
        """
        parameters_to_test = [
            # OpenEdx
            "lis_person_sourcedid",
            # Moodle
            "ext_user_username",
        ]
        for param in parameters_to_test:
            user_id = lti_request.get_param(param)
            if user_id:
                return user_id
        logger.debug("Unable to find remote user id in LTI request")
        raise PermissionDenied()



"""Views of the lti_provider django application."""
from abc import ABC, abstractmethod

from django import http
from django.contrib.auth import authenticate, login
from django.http import HttpRequest, HttpResponse, HttpResponseForbidden
from django.utils.decorators import method_decorator
from django.utils.module_loading import import_string
from django.views import View
from django.views.decorators.csrf import csrf_exempt

from lti_provider.exceptions import LTIException

from .defaults import LTI_LAUNCH_FAILURE_HANDLER, LTI_LAUNCH_SUCCESS_HANDLER
from .lti import LTI


@csrf_exempt
def launch(request: HttpRequest) -> HttpResponse:
    """Verify LTI launch request and call hook depending on the result."""
    lti = LTI(request)
    try:
        lti.verify()
        response = import_string(LTI_LAUNCH_SUCCESS_HANDLER)(request, lti)
        # If no response is given by the callback, send an empty HTTP 200 response
        return (
            response if isinstance(response, http.HttpResponse) else http.HttpResponse()
        )
    except LTIException as error:
        response = import_string(LTI_LAUNCH_FAILURE_HANDLER)(request, error)
        if isinstance(response, http.HttpResponse):
            return response
        # If no response is given by the callback, send an empty HTTP 403 response
        return http.HttpResponseForbidden()


@method_decorator(csrf_exempt, name="dispatch")
class BaseLTIView(ABC, View):
    """
    Abstract view to help building LTI authenticated views.
    It is convenient if you want to handle dynamic LTI launch URLs.
    """

    def post(self, request, *args, **kwargs) -> HttpResponse:  # pylint: disable=W0613
        """Handler for POST requests."""
        lti_request = LTI(request)
        try:
            lti_request.verify()
            user = authenticate(request, lti_request=lti_request)
            if user is not None:
                login(request, user)
                return self._do_on_login(lti_request)
            return HttpResponseForbidden()
        except LTIException:
            return HttpResponseForbidden()

    @abstractmethod
    def _do_on_login(self, lti_request: LTI) -> HttpResponse:
        """Process the request when the user is logged in via LTI"""
        raise NotImplementedError()



"""This module contains default settings for lti_provider application."""

from django.conf import settings

# When a LTI launch request is successfully verified, the handler function defined
# in the LTI_LAUNCH_SUCCESS_HANDLER setting is called with 2 parameters :
# 1. The request (django.http.request.HttpRequest)
# 2. The verified LTI request (lti_provider.lti.LTI)
#
# The handler should return a valid HTTP response (django.http.HttpResponse)
# A default implementation is given in this application as an exemple
#
# If you use the default implementation, you must add the following
# authentication backend in your AUTHENTICATION_BACKENDS setting:
#   "lti_provider.default.backend.AuthBackend"
#

LTI_LAUNCH_SUCCESS_HANDLER = getattr(
    settings, "LTI_LAUNCH_SUCCESS_HANDLER", "lti_provider.default.handlers.success"
)

# When a LTI launch request is invalid, the handler function defined in the
# LTI_LAUNCH_FAILURE_HANDLER is called with the following parameters:
# 1. The request (django.http.request.HttpRequest)
# 2. The LTIException
#
# The handler should return a valid HTTP response (django.http.HttpResponse)
# A default implementation is given in this application as an example.
LTI_LAUNCH_FAILURE_HANDLER = getattr(
    settings, "LTI_LAUNCH_FAILURE_HANDLER", "lti_provider.default.handlers.failure"
)



"""Factories for the ``ashley`` application."""
import factory
from factory.django import DjangoModelFactory
from machina.core.db.models import get_model

from lti_provider.factories import LTIConsumerFactory

LTIContext = get_model("ashley", "LTIContext")  # pylint: disable=C0103
User = get_model("ashley", "User")  # pylint: disable=C0103


class UserFactory(DjangoModelFactory):
    """Factory to create a User."""

    class Meta:
        model = User

    lti_consumer = factory.SubFactory(LTIConsumerFactory)

    public_username = factory.Sequence(lambda n: f"user{n}")
    email = factory.LazyAttribute(lambda u: f"{u.public_username}@example.com")
    lti_remote_user_id = factory.SelfAttribute("public_username")
    username = factory.LazyAttribute(
        lambda u: f"{u.public_username}@{u.lti_consumer.slug}"
    )


class LTIContextFactory(DjangoModelFactory):
    """Factory to create a LTIContext."""

    class Meta:
        model = LTIContext

    lti_consumer = factory.SubFactory(LTIConsumerFactory)
    lti_id = factory.Sequence(lambda n: f"context{n}")



"""This module contains default handlers for LTI launch request."""

from django.contrib.auth import authenticate, login
from django.http import HttpRequest, HttpResponse, HttpResponseForbidden

from ..exceptions import LTIException
from ..lti import LTI


def success(request: HttpRequest, lti_request: LTI) -> HttpResponse:
    """
    Default handler for successfully verified LTI launch requests.
    You are encouraged to define your own handler according to your project needs.
    It can be used to call an authentication backend, to redirect the user...etc

    See LTI_LAUNCH_SUCCESS_HANDLER setting.

    Args:
        request: The HTTP request
        lti_request: The verified LTI request

    Returns:
        HttpResponse: An HTTP response
    """

    user = authenticate(request, lti_request=lti_request)
    if user is not None:
        login(request, user)
        return HttpResponse("Welcome, user {}".format(user.username))
    return HttpResponseForbidden("Permission denied")


# pylint: disable=unused-argument
def failure(request: HttpRequest, error: LTIException) -> HttpResponse:
    """
    Default handler for invalid LTI launch requests.
    You are encouraged to define your own handler according to your project needs.

    See LTI_LAUNCH_FAILURE_HANDLER setting.

    Args:
        request: the HTTP request
        error: The LTIException encountered

    Returns:
        HttpResponse: An HTTP response
    """
    return HttpResponseForbidden("Invalid LTI launch request")



"""Declare the middlewares related to Ashley ."""

from django.conf import settings
from django.utils.deprecation import MiddlewareMixin


class SameSiteNoneMiddleware(MiddlewareMixin):
    """
    This middleware forces the attribute SameSite="None" for CSRF and SESSION
    cookies.

    Django < 3.1 does not allow to set the SameSite="None" for the
    session and CSRF cookies. As a result, it causes issues with modern
    browsers that consider SameSite="Lax" by default.

    This middleware is a temporary fix until this feature is available
    in a stable version of Django.
    """

    @staticmethod
    def process_response(request, response):
        """Force the samesite=None on session and CSRF cookies if defined."""

        if settings.SESSION_COOKIE_NAME in response.cookies:
            response.cookies[settings.SESSION_COOKIE_NAME]["samesite"] = "None"
        if settings.CSRF_COOKIE_NAME in response.cookies:
            response.cookies[settings.CSRF_COOKIE_NAME]["samesite"] = "None"
        return response



"""This module contains an authentication backend based on LTI launch request."""

import logging
from typing import Any

from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.core.exceptions import PermissionDenied

from ..lti import LTI

logger = logging.getLogger(__name__)

USER_MODEL = get_user_model()


class LTIBackend(ModelBackend):
    """
    Authentication backend used by the lti_provider.default.handlers.success handler.
    It authenticates a user from a verified LTI request and creates a User if necessary.

    You are encouraged to make your own authentication backend to add your own domain logic.
    """

    def authenticate(self, request, username=None, password=None, **kwargs):
        """
        Authenticate a user from a LTI request

        Args:
            request: django http request
            username: The (optional) username to authenticate
            password: The (optional) password of the user to authenticate
            kwargs: additional parameters

        Returns:
            An authenticated user or None
        """

        lti_request = kwargs.get("lti_request")
        if not lti_request:
            return None

        if not lti_request.is_valid:
            raise PermissionDenied()

        username = self._get_mandatory_param(lti_request, "user_id")
        email = self._get_mandatory_param(
            lti_request, "lis_person_contact_email_primary"
        )

        logger.debug("User %s authenticated from LTI request", username)

        try:
            user = USER_MODEL.objects.get_by_natural_key(username)
        except USER_MODEL.DoesNotExist:
            user = USER_MODEL.objects.create_user(username, email=email)
            logger.debug("User %s created in database", username)
        if not user.is_active:
            logger.debug("User %s is not active", user.username)
            raise PermissionDenied()
        return user

    @staticmethod
    def _get_mandatory_param(lti_request: LTI, param: str) -> Any:
        """Get a LTI parameter or throw a exception if not defined."

        Args:
            lti_request: The verified LTI request
            param: A LTI parameter name

        Returns: The parameter value

        Raises:
            PermissionDenied if the parameter is not defined
        """
        value = lti_request.get_param(param)
        if not value:
            logger.debug("Unable to find param %s in LTI request", param)
            raise PermissionDenied()
        logger.debug("%s = %s", param, value)
        return value



"""This module contains django widgets related to Ashley WYSIWYG editor """

import json

from django.forms.widgets import HiddenInput
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _


class DraftEditor(HiddenInput):
    """Draft.js editor form field widget."""

    template_name = "widgets/editor.html"
    props = {
        "placeholder": _("Please enter your message here..."),
    }

    def __init__(self, attrs=None, props=None):
        """Allow optional extension of widget props in init."""
        super().__init__(attrs)
        if props:
            self.props.update(**props)

    def get_context(self, name, value, attrs):
        """Add props to context."""
        context = super().get_context(name, value, attrs)
        context["widget"]["props"] = mark_safe(json.dumps(self.props))  # nosec
        return context


__all__ = ["DraftEditor"]



"""Views of the ashley django application."""

import logging
from typing import List

from django.conf import settings
from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseRedirect
from django.urls import reverse
from django.utils import translation
from machina.apps.forum_permission.shortcuts import assign_perm
from machina.core.db.models import get_model

from lti_provider.lti import LTI
from lti_provider.views import BaseLTIView

from .defaults import DEFAULT_FORUM_BASE_PERMISSIONS, DEFAULT_FORUM_ROLES_PERMISSIONS

Forum = get_model("forum", "Forum")  # pylint: disable=C0103
LTIContext = get_model("ashley", "LTIContext")  # pylint: disable=C0103

logger = logging.getLogger(__name__)


class ForumLTIView(BaseLTIView):
    """Forum view called by an LTI launch request."""

    def _do_on_login(self, lti_request: LTI) -> HttpResponse:
        """Process the request when the user is logged in via LTI"""

        # Get or create the LTIContext model associated with the current LTI launch request
        context_id = lti_request.get_param("context_id")
        if not context_id:
            return HttpResponseBadRequest("LTI parameter context_id is mandatory")
        context, _context_created = LTIContext.objects.get_or_create(
            lti_id=context_id, lti_consumer=lti_request.get_consumer()
        )

        # Synchronize the user groups related to the current LTI context
        context.sync_user_groups(self.request.user, lti_request.roles)

        forum_uuid = self.kwargs["uuid"]
        forum, forum_created = Forum.objects.get_or_create(
            lti_id=forum_uuid,
            type=Forum.FORUM_POST,
            defaults={"name": lti_request.context_title},
        )
        if forum_created:
            logger.debug("Forum %s created", forum_uuid)
            self._init_forum(forum, context)

        response = HttpResponseRedirect(
            reverse("forum:forum", kwargs={"slug": forum.slug, "pk": forum.id})
        )

        locale = lti_request.get_param("launch_presentation_locale")
        if locale:
            logger.debug("Course locale detected %s", locale)
            translation.activate(locale)
            response.set_cookie(settings.LANGUAGE_COOKIE_NAME, locale)

        return response

    def _init_forum(self, forum, context):
        """
        Initialize a forum.

        This method sets default permissions according to the LTI context that
        initiated the creation of the forum.
        """
        forum.lti_contexts.add(context)
        self._assign_permissions(
            forum, context.get_base_group(), DEFAULT_FORUM_BASE_PERMISSIONS,
        )
        for role, perms in DEFAULT_FORUM_ROLES_PERMISSIONS.items():
            self._assign_permissions(forum, context.get_role_group(role), perms)

    @staticmethod
    def _assign_permissions(forum, group, permissions: List[str]):
        """Grant a list of permissions for a group on a specific forum."""
        for perm in permissions:
            logger.debug(
                "Grant permission %s to group %s on forum %s",
                perm,
                group.name,
                forum.lti_id,
            )
            assign_perm(perm, group, forum, True)



"""lti_provider application."""
from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class LtiProviderAppConfig(AppConfig):
    """Configuration class for the lti_provider app."""

    verbose_name = _("LTI Provider")
    name = "lti_provider"



"""This module contains default settings for Ashley application."""

from django.conf import settings
from draftjs_exporter.defaults import BLOCK_MAP as DEFAULT_BLOCK_MAP
from draftjs_exporter.defaults import STYLE_MAP as DEFAULT_STYLE_MAP
from draftjs_exporter.dom import DOM

from ashley.editor.decorators import emoji, link

_FORUM_BASE_PERMISSIONS = [
    "can_see_forum",
    "can_read_forum",
    "can_start_new_topics",
    "can_reply_to_topics",
    "can_edit_own_posts",
    "can_post_without_approval",
    "can_vote_in_polls",
]

_FORUM_ADMIN_PERMISSIONS = _FORUM_BASE_PERMISSIONS + [
    "can_post_announcements",
    "can_post_stickies",
    "can_delete_own_posts",
    "can_create_polls",
    "can_lock_topics",
    "can_edit_posts",
    "can_delete_posts",
    "can_approve_posts",
    "can_reply_to_locked_topics",
]

DEFAULT_FORUM_BASE_PERMISSIONS = getattr(
    settings, "ASHLEY_DEFAULT_FORUM_BASE_PERMISSIONS", _FORUM_BASE_PERMISSIONS
)

DEFAULT_FORUM_ROLES_PERMISSIONS = getattr(
    settings,
    "ASHLEY_DEFAULT_FORUM_ROLES_PERMISSIONS",
    {
        "administrator": _FORUM_ADMIN_PERMISSIONS,
        "instructor": _FORUM_ADMIN_PERMISSIONS,
    },
)


DEFAULT_DRAFTJS_EXPORTER_CONFIG = {
    "entity_decorators": {"LINK": link, "emoji": emoji},
    "composite_decorators": [],
    "block_map": DEFAULT_BLOCK_MAP,
    "style_map": DEFAULT_STYLE_MAP,
    "engine": DOM.STRING,
}

DRAFTJS_EXPORTER_CONFIG = {
    **DEFAULT_DRAFTJS_EXPORTER_CONFIG,
    **getattr(settings, "ASHLEY_DRAFTJS_EXPORTER_CONFIG", {}),
}



# Generated by Django 2.2.10 on 2020-02-26 18:25

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="LTIConsumer",
            fields=[
                (
                    "slug",
                    models.SlugField(
                        help_text="identifier for the consumer site",
                        primary_key=True,
                        serialize=False,
                        verbose_name="consumer site identifier",
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        help_text="human readable title, to describe the LTI consumer",
                        max_length=255,
                        verbose_name="Title",
                    ),
                ),
            ],
            options={
                "verbose_name": "LTI consumer",
                "verbose_name_plural": "LTI consumers",
                "db_table": "lti_consumer",
            },
        ),
        migrations.CreateModel(
            name="LTIPassport",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        help_text="human readable title, to describe this LTI passport (i.e. : who will use it ?)",
                        max_length=255,
                        verbose_name="Title",
                    ),
                ),
                (
                    "oauth_consumer_key",
                    models.CharField(
                        editable=False,
                        help_text="oauth consumer key to authenticate an LTI consumer on the LTI provider",
                        max_length=255,
                        unique=True,
                        verbose_name="oauth consumer key",
                    ),
                ),
                (
                    "shared_secret",
                    models.CharField(
                        editable=False,
                        help_text="LTI Shared secret",
                        max_length=255,
                        verbose_name="shared secret",
                    ),
                ),
                (
                    "is_enabled",
                    models.BooleanField(
                        default=True,
                        help_text="whether the passport is enabled",
                        verbose_name="is enabled",
                    ),
                ),
                (
                    "consumer",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="lti_provider.LTIConsumer",
                    ),
                ),
            ],
            options={
                "verbose_name": "LTI passport",
                "verbose_name_plural": "LTI passports",
                "db_table": "lti_passport",
            },
        ),
    ]



"""This module contains custom decorators for draftjs_exporter."""
import logging

from draftjs_exporter.dom import DOM

logger = logging.getLogger(__name__)


def link(props):
    """
    Decorator for the `LINK` entity in Draft.js ContentState.

    `draftjs_exporter` does not provide a decorator for this kind of entity by
     default, so we have to define it to support links in forum posts.
    """
    title = props.get("title")
    href = props.get("url", "#")
    anchor_properties = {
        "href": href,
        "target": "_blank",
        "rel": "nofollow noopener noreferrer",
    }

    if title is not None:
        anchor_properties["title"] = title

    return DOM.create_element("a", anchor_properties, props["children"])


def emoji(props):
    """
    Decorator for the `emoji` entity in Draft.js ContentState.

    This entity is added by the `draft-js-emoji-plugin` plugin.
    (https://www.draft-js-plugins.com/plugin/emoji)
    """
    return DOM.create_element("span", {"class": "emoji"}, props["children"])



"""Admin of the lti_provider application."""

from django.contrib import admin

from .models import LTIConsumer, LTIPassport


@admin.register(LTIConsumer)
class LTIConsumerAdmin(admin.ModelAdmin):
    """Admin class for the LTIConsumer model."""

    list_display = (
        "slug",
        "title",
    )


@admin.register(LTIPassport)
class LTIPassportAdmin(admin.ModelAdmin):
    """Admin class for the LTIPassport model."""

    list_display = (
        "title",
        "oauth_consumer_key",
        "is_enabled",
    )

    readonly_fields = [
        "oauth_consumer_key",
        "shared_secret",
    ]
